/***********************************************************************

(C) Copyright 2009, Vadim von Brzeski.  vadimv@numberjack.net

dp_gibbs is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dp_gibbs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dp_gibbs.  If not, see <http://www.gnu.org/licenses/>.

************************************************************************/

#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_errno.h>
#include <iostream>
#include <fstream>
#include <math.h>
#include <vector>
#include <map>
#include <string>
#include <boost/tokenizer.hpp>
#include <gflags/gflags.h>

DEFINE_string(mode, "est", "'est' to run gibbs sampler, 'rest' to re-run gibbs sampler using new docs and prior model, or 'inf' to infer topic distribution of new docs using existing model");
DEFINE_string(datafile, "", "the input data matrix, N documents x M words");
DEFINE_string(modeldir, "/tmp", "the pathname to the directory for model files");
DEFINE_string(outdir, "/tmp", "the pathname to the directory for output files");
DEFINE_int32(iterations, 10, "total number of gibbs sampling iterations");
DEFINE_int32(burn, 5, "number of burn in iterations");
DEFINE_double(beta, 1.0, "hyperparameter for theta ~ G(0) = Dirichlet(beta), symmetric DP base distribution; larger beta implies fewer clusters.");
DEFINE_double(a, 0.1, "hyperparameter for alpha ~ Gamma(a,b)");
DEFINE_double(b, 0.1, "hyperparameter for alpha ~ Gamma(a,b)");

void getLogLikelihoodExistingClusters(long double * LL, const gsl_matrix * X, const gsl_matrix * Q, const double * sum_Q, size_t doc_i, size_t K, size_t M, double beta);
long double getLogLikelihoodNewCluster(const gsl_matrix * X, size_t doc_i, size_t M, double beta);
size_t sampleFromClusters(long double * LL_existing, long double LL_new, const gsl_matrix * clusters, size_t doc_i, size_t K, double alpha, const double * sum_clusters);
void estimateViaGibbs(const gsl_matrix * X, size_t N, size_t M, const double * sum_X, gsl_matrix *& clusters, double *& sum_clusters, size_t& K,
                        gsl_matrix *& Q, double *& sum_Q,
                        std::map<size_t, std::map<size_t, int> >& sampled_clusters, std::map<size_t, int>& sampled_K,
                        double beta, double& alpha, double a, double b, size_t nIterations, size_t nBurnIn);
void ehandler (const char * reason, const char * file, int line, int gsl_errno);

// utility and debug routines
bool sum(const gsl_matrix * A, size_t nRows, size_t nCols, size_t dim, double * result);
void normalize(gsl_matrix * A, size_t nRows, size_t nCols, size_t dim);
void gammaln(const gsl_matrix * A, size_t nRows, size_t nCols, gsl_matrix * Out, double c = 0);
void gammaln(const gsl_vector * V, size_t dim, gsl_vector * Out, double c = 0);
bool readCSVMatrix(const std::string& filename, gsl_matrix * A, size_t nRows, size_t nCols);
bool readSparseMatrix(const std::string& filename, gsl_matrix *& A, size_t& nRows, size_t& nCols);
void writeSparseMatrix(const std::string& filename, const gsl_matrix * A, size_t nRows, size_t nCols);
void writeOutput(const std::string& outdir, std::map<size_t, std::map<size_t, int> >& sampled_clusters, const std::map<size_t, int>& sampled_K,
                const gsl_matrix * X, gsl_matrix * Q, const gsl_matrix * clusters, size_t N, size_t M, size_t K, double alpha);
bool readModel(const std::string& modeldir, size_t& N, size_t& M, gsl_matrix *& clusters, double *& sum_clusters, size_t& K, gsl_matrix *& Q, double *& sum_Q, gsl_matrix *& Qnorm,
                std::map<size_t, std::map<size_t, int> >& sampled_clusters, std::map<size_t, int>& sampled_K, double& alpha);
void dump(const gsl_vector * V, size_t dim);
void dump(const double * V, size_t dim);
void dumpl(const long double * V, size_t dim);
void dump(const gsl_matrix * X, size_t N, size_t M);


int main(int argc, char** argv)
{
        if (argc < 2)
        {
                std::string usage = "\nCollapsed Gibbs sampler for Dirichlet Process model.  Based on \n\n\"Modeling individual differences using Dirichlet processes\", ";
                usage += "Navarro, Griffiths, Steyvers, Lee, 2005.\n\n";
                usage += "Usage : dp_gibbs -mode=[est|rest|inf] -datafile=<doc-word matrix file> -modeldir=<model directory path> -outdir=<output dir> ";
                usage += "-iterations=<num> -burn=<num> -beta=<beta> -a=<a> -b<b>\n\n";
                usage += "\ndp_gibbs --helpshort explains the options in more detail\n\n";
                std::cerr << usage;
                google::SetUsageMessage("");
                google::ParseCommandLineFlags(&argc, &argv, true);
                exit(0);
        }

        google::ParseCommandLineFlags(&argc, &argv, true);

        gsl_set_error_handler (&ehandler);

        std::string mode = FLAGS_mode;

        size_t N; // number of data points
        size_t M; // number of unique responses per data point

        size_t nIterations = FLAGS_iterations;
        size_t nBurnIn = FLAGS_burn;

        // initial alpha value - will be sampled below
        double alpha = 1.0;

        // hyperparameters for alpha ~ Gamma(a,b)
        double a = FLAGS_a;
        double b = FLAGS_b;

        // hyperparameter for theta ~ G(0) = Dirichlet(beta)  (symmetric DP base distribution)
        // larger beta implies less distinct (peaked) multinomial distributions over the M responses - less distinct topics --> fewer clusters
        // smaller beta implies more distinct (peaked) multinomial distn's over M responses - more distinct topics --> more clusters
        double beta = FLAGS_beta;

        // data matrix : data instance (doc) x response (word)
        gsl_matrix * X = NULL;
        std::string filename(FLAGS_datafile);
        if (!readSparseMatrix(FLAGS_datafile, X, N, M))
        {
                std::cerr << "Could not read data matrix " << FLAGS_datafile << "\n";
                exit(0);
        }
        // keep track of sum of words in each document
        double * sum_X = new double[N];
        sum(X, N, M, 2, sum_X);

        // cluster assignment matrix : cluster (index) x data instance (doc)
        // clusters(i,j) = 1 iff data instance j is in cluster i
        gsl_matrix * clusters = gsl_matrix_alloc(1, N);
        gsl_matrix_set_zero(clusters);
        double * sum_clusters = new double[1];
        sum_clusters[0] = 0;
        // keeps track of which clusters a doc was assigned to over the sampling iterations
        std::map<size_t, std::map<size_t, int> > sampled_clusters;
        for (size_t d = 0; d < N; d++) sampled_clusters[d][0] = 0;
        // keeps track of sampled number of clusters
        std::map<size_t, int> sampled_K;

        // keep track of number of clusters
        size_t K = 0;

        // matrix : cluster (index) x response (word)
        // Q = clusters * X;
        gsl_matrix * Q = gsl_matrix_alloc(1, M);
        gsl_matrix * Qnorm = NULL;
        gsl_matrix_set_zero(Q);
        double * sum_Q = new double[1];
        sum_Q[0] = 0;

        // run full gibbs sampling from scratch on all documents
        if (mode == "est")
        {
                estimateViaGibbs(X, N, M, sum_X, clusters, sum_clusters, K, Q, sum_Q, sampled_clusters, sampled_K, beta, alpha, a, b, nIterations, nBurnIn);
                writeOutput(FLAGS_outdir, sampled_clusters, sampled_K, X, Q, clusters, N, M, K, alpha);
        }
        // run gibbs sampling incorporating additional document(s), but using existing topic-model from previous estimationA
        // (approximation)
        else if (mode == "rest")
        {
                size_t previous_N;              // the number of docs used to build the initial model
                if (readModel(FLAGS_modeldir, previous_N, M, clusters, sum_clusters, K, Q, sum_Q, Qnorm, sampled_clusters, sampled_K, alpha))
                {
                        // allocate a new cluster member matrix for new docs
                        gsl_matrix_free(clusters);
                        clusters = gsl_matrix_alloc(K, N);
                        gsl_matrix_set_zero(clusters);

                        // clear previous sampled statistics
                        sampled_clusters.clear();
                        sampled_K.clear();

                        // run gibbs sampling, but with 0 burn in
                        estimateViaGibbs(X, N, M, sum_X, clusters, sum_clusters, K, Q, sum_Q, sampled_clusters, sampled_K, beta, alpha, a, b, nIterations, 0);

                        writeOutput(FLAGS_outdir, sampled_clusters, sampled_K, X, Q, clusters, N, M, K, alpha);

                }
        }
        // estimate the topic (cluster) membership of new doc x using existing topic model (w/o gibbs), i.e. inf = X*transpose(Q).
        else if (mode == "inf")
        {
                size_t previous_N;              // the number of docs used to build the initial model
                if (readModel(FLAGS_modeldir, previous_N, M, clusters, sum_clusters, K, Q, sum_Q, Qnorm, sampled_clusters, sampled_K, alpha))
                {
                        gsl_matrix * inferredClusters = gsl_matrix_alloc(N, K);
                        // matrix multiply
                        gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, X, Qnorm, 0.0, inferredClusters);
                        // normalize
                        normalize(inferredClusters, N, K, 2);
                        writeSparseMatrix(FLAGS_outdir+"/X_inf.txt", inferredClusters, N, K);
                        gsl_matrix_free(inferredClusters);
                }
        }

        // free memory
        if (clusters != NULL) gsl_matrix_free(clusters);
        if (X != NULL) gsl_matrix_free(X);
        delete [] sum_X;
        delete [] sum_Q;
        delete [] sum_clusters;
        if (Q != NULL) gsl_matrix_free(Q);
        if (Qnorm != NULL) gsl_matrix_free(Qnorm);

}

void estimateViaGibbs(const gsl_matrix * X, size_t N, size_t M, const double * sum_X, gsl_matrix *& clusters, double *& sum_clusters, size_t& K, 
			gsl_matrix *& Q, double *& sum_Q, std::map<size_t, std::map<size_t, int> >& sampled_clusters, 
			std::map<size_t, int>& sampled_K, double beta, double& alpha, double a, double b, size_t nIterations, size_t nBurnIn)
{

        // setup random number generator
        const gsl_rng_type * T;
        gsl_rng * r;
        /* create a generator chosen by the environment variable GSL_RNG_TYPE */
        gsl_rng_env_setup();
        T = gsl_rng_default;
        r = gsl_rng_alloc (T);
        srand(time(NULL));
        gsl_rng_set (r, rand());

        size_t * permutation = new size_t[N];
        for (size_t t = 0; t < N; t++) permutation[t] = t;

        for (int iter = 1; iter <= nIterations; iter++)
        {

                // need to do the random permutation of N data samples
                gsl_rng_set (r, rand());
                gsl_ran_shuffle (r, permutation, N, sizeof(size_t));

                for (size_t t = 0; t < N; t++)
                {
                        size_t doc_i = permutation[t];

                        // need to handle special case when K = 0
                        long double * likelihood_existing_clusters = NULL;
                        size_t z_i = 0;
                        if (K > 0)
                        {
                                likelihood_existing_clusters = new long double[K];
                                getLogLikelihoodExistingClusters(likelihood_existing_clusters, X, Q, sum_Q, doc_i, K, M, beta);

                                long double likelihood_new_cluster = getLogLikelihoodNewCluster(X, doc_i, M, beta);

                                z_i = sampleFromClusters(likelihood_existing_clusters, likelihood_new_cluster, clusters, doc_i, K, alpha, sum_clusters);
                        }

                        // update Q and clusters matrices

                        if (z_i >= K && K > 0)
                        {
                                // add one more row to the clusters and Q matrices, and sum_Q and sum_clusters vectors

                                gsl_matrix * _temp = gsl_matrix_alloc(K+1, N);
                                gsl_matrix_set_zero(_temp);
                                for (size_t j = 0; j < K; j++) gsl_matrix_set_row (_temp, j, &(gsl_matrix_const_row(clusters, j)).vector);
                                gsl_matrix_free(clusters);
                                clusters = _temp;

                                _temp = gsl_matrix_alloc(K+1, M);
                                gsl_matrix_set_zero(_temp);
                                for (size_t j = 0; j < K; j++) gsl_matrix_set_row (_temp, j, &(gsl_matrix_const_row(Q, j)).vector);
                                gsl_matrix_free(Q);
                                Q = _temp;

                                double * _sum_temp = new double[K+1];
                                for (size_t j = 0; j < K; j++) _sum_temp[j] = sum_Q[j];
                                _sum_temp[K] = 0;
                                delete [] sum_Q;
                                sum_Q = _sum_temp;

                                double * _sum_c = new double[K+1];
                                for (size_t j = 0; j < K; j++) _sum_c[j] = sum_clusters[j];
                                _sum_c[K] = 0;
                                delete [] sum_clusters;
                                sum_clusters = _sum_c;

                                // for now, increment K
                                K += 1;
                        }
                        if (K == 0) K = 1;

                        size_t prev_cluster_for_doc_i = gsl_vector_max_index(&(gsl_matrix_const_column(clusters,doc_i)).vector);
                        // check if there really was a 1 at that index or if it was one of many 0 values
                        if (gsl_matrix_get(clusters, prev_cluster_for_doc_i, doc_i) == 0)
                        {
                                gsl_matrix_set(clusters, z_i, doc_i, 1);
                                sum_clusters[z_i] += 1;
                                // update Q statistics
                                // Q(z_i,:) = Q(z_i,:) + X(doc_i,:);
                                gsl_vector_add(&(gsl_matrix_row(Q, z_i)).vector, &(gsl_matrix_const_row(X, doc_i)).vector);
                                sum_Q[z_i] += sum_X[doc_i];
                        }
                        else if (prev_cluster_for_doc_i != z_i)
                        {
                                // assign doc_i to new cluster z_i and zero-out old cluster
                                gsl_matrix_set(clusters, prev_cluster_for_doc_i, doc_i, 0);
                                sum_clusters[prev_cluster_for_doc_i] -= 1;
                                gsl_matrix_set(clusters, z_i, doc_i, 1);
                                sum_clusters[z_i] += 1;
                                // update Q statistics
                                // Q(z_i,:) = Q(z_i,:) + X(doc_i,:);
                                gsl_vector_add(&(gsl_matrix_row(Q, z_i)).vector, &(gsl_matrix_const_row(X, doc_i)).vector);
                                sum_Q[z_i] += sum_X[doc_i];
                                // Q(prev_z_i,:) = Q(prev_z_i, :) - X(doc_i,:);
                                gsl_vector_sub(&(gsl_matrix_row(Q, prev_cluster_for_doc_i)).vector, &(gsl_matrix_const_row(X, doc_i)).vector);
                                sum_Q[prev_cluster_for_doc_i] -= sum_X[doc_i];
                        }

                        // removeEmptyClusters(Q, clusters, sum_Q, sum_clusters)
                        if (K > 1)
                        {
                                int emptyIndex = -1;
                                for (size_t z = 0; z < K; z++)
                                {
                                        // if (gsl_vector_isnull(&(gsl_matrix_const_row(clusters,z)).vector))
                                        if (sum_clusters[z] == 0)
                                        {
                                                emptyIndex = z;
                                                break;
                                        }
                                }
                                if (emptyIndex >= 0)
                                {

                                        gsl_matrix * _tempClusters = gsl_matrix_alloc(K-1, N);
                                        gsl_matrix * _tempQ = gsl_matrix_alloc(K-1, M);
                                        double * _temp_sum = new double[K-1];
                                        double * _temp_c = new double[K-1];

                                        for (size_t z = 0; z < K; z++)
                                        {
                                                if (z < emptyIndex)
                                                {
                                                        gsl_matrix_set_row(_tempClusters, z, &(gsl_matrix_const_row(clusters, z)).vector);
                                                        gsl_matrix_set_row(_tempQ, z, &(gsl_matrix_const_row(Q, z)).vector);
                                                        _temp_sum[z] = sum_Q[z];
                                                        _temp_c[z] = sum_clusters[z];
                                                }
                                                else if (z > emptyIndex)
                                                {
                                                        gsl_matrix_set_row(_tempClusters, z - 1, &(gsl_matrix_const_row(clusters, z)).vector);
                                                        gsl_matrix_set_row(_tempQ, z - 1, &(gsl_matrix_const_row(Q, z)).vector);
                                                        _temp_sum[z-1] = sum_Q[z];
                                                        _temp_c[z-1] = sum_clusters[z];
                                                }
                                        }

                                        gsl_matrix_free(clusters);
                                        clusters = _tempClusters;

                                        gsl_matrix_free(Q);
                                        Q = _tempQ;

                                        delete [] sum_Q;
                                        sum_Q = _temp_sum;

                                        delete [] sum_clusters;
                                        sum_clusters = _temp_c;

                                        K -= 1;

                                }
                        }

                        delete [] likelihood_existing_clusters;

                } // end for each document

                // sample alpha
                gsl_rng_set (r, rand());
                double eta = gsl_ran_beta(r, alpha, N);
                if (eta > 0)
                        alpha = gsl_ran_gamma(r, (double)(a + K), (double)(b - log(eta)));
                else
                        alpha = 1.0;

                // keep track of sampled cluster assignments
                if (iter > nBurnIn)
                {
                        for (size_t d = 0; d < N; d++)
                        {

                                gsl_vector_view _col = gsl_matrix_column(clusters, d);
                                size_t z = gsl_vector_max_index(&(_col).vector);
                                if (sampled_clusters[d].find(z) == sampled_clusters[d].end()) sampled_clusters[d][z] = 1;
                                else sampled_clusters[d][z] += 1;
                        }
                }

                std::cerr << "Iteration : " << iter << " / " << nIterations << ".  K = " << K << "\n";

                // keep track of sampled number of clusters
                if (sampled_K.find(K) == sampled_K.end()) sampled_K[K] = 1;
                else sampled_K[K] += 1;

        }  // end sampling iteration

}


bool readModel(const std::string& modeldir, size_t& N, size_t& M, gsl_matrix *& clusters, double *& sum_clusters, size_t& K, 
		gsl_matrix *& Q, double *& sum_Q, gsl_matrix *& Qnorm,
                std::map<size_t, std::map<size_t, int> >& sampled_clusters, std::map<size_t, int>& sampled_K, double& alpha)
{
        if (Q != NULL) gsl_matrix_free(Q);
        std::string filename("Q");
        bool ret = readSparseMatrix(modeldir+"/"+filename, Q, K, M);
        if (!ret)
        {
                std::cerr << "Error : could not read model : Q matrix\n";
                return false;
        }

        delete [] sum_Q;
        sum_Q = new double[K];
        sum(Q, K, M, 2, sum_Q);

        if (Qnorm != NULL) gsl_matrix_free(Qnorm);
        filename = "Qnorm_c";
        if (!readSparseMatrix(modeldir+"/"+filename, Qnorm, K, M))
        {
                std::cerr << "Error : could not read model : Qnorm_c matrix\n";
                return false;
        }

        if (clusters != NULL) gsl_matrix_free(clusters);
        filename = "clusters.final";
        if (!readSparseMatrix(modeldir+"/"+filename, clusters, K, N))
        {
                std::cerr << "Error : could not read model : clusters matrix\n";
                return false;
        }

        delete [] sum_clusters;
        sum_clusters = new double[K];
        sum(clusters, K, N, 2, sum_clusters);

        std::ifstream afile((modeldir+"/alpha").c_str());
        afile >> alpha;
        afile.close();

        return true;
}


void writeSparseMatrix(const std::string& filename, const gsl_matrix * A, size_t nRows, size_t nCols)
{
        std::ofstream cOut(filename.c_str());
        cOut << nRows << " " << nCols << "\n";
        for (size_t i = 0; i < nRows; i++)
        {
                cOut << i << " ";
                for (size_t j = 0; j < nCols-1; j++)
                {
                        cOut << j << ":" << gsl_matrix_get(A, i, j) << " ";
                }
                cOut << (nCols-1) << ":" << gsl_matrix_get(A, i, nCols-1) << "\n";
        }
        cOut.close();
}


void writeOutput(const std::string& outdir, std::map<size_t, std::map<size_t, int> >& sampled_clusters, const std::map<size_t, int>& sampled_K,
                const gsl_matrix * X, gsl_matrix * Q, const gsl_matrix * clusters, size_t N, size_t M, size_t K, double alpha)
{

        // write out sampled cluster memberships over the iterations
        std::ofstream clustersOut((outdir+"/clusters.dist").c_str());
        clustersOut << N << " " << K << "\n";
        for (size_t d = 0; d < N; d++)
        {
                double s = 0.0;
                for (size_t k = 0; k < K; k++) s += sampled_clusters[d][k];
                for (size_t k = 0; k < K; k++)
                {
                        clustersOut << k << ":" << (double)sampled_clusters[d][k]/s << " ";
                }
                clustersOut << "\n";
        }
        clustersOut.close();

        // write out the final cluster membership matrix
        writeSparseMatrix(outdir+"/clusters.final", clusters, K, N);

        // write out sampled number of clusters values
        std::ofstream kOut((outdir+"/K_histogram").c_str());
        for (std::map<size_t, int>::const_iterator it = sampled_K.begin(); it != sampled_K.end(); ++it)
        {
                kOut << it->first << ":" << it->second << " ";
        }
        kOut.close();

        // write out Q matrix
        writeSparseMatrix(outdir+"/Q", Q, K, M);

        gsl_matrix * Qr = gsl_matrix_alloc(K, M);
        gsl_matrix_memcpy(Qr, Q);

        // normalize Q by topics to get p(topic_k|word)
        normalize(Q, K, M, 1);
        writeSparseMatrix(outdir+"/Qnorm_c", Q, K, M);

        // normalize Qr by words to get p(word|topic_k)
        normalize(Qr, K, M, 2);
        writeSparseMatrix(outdir+"/Qnorm_r", Qr, K, M);
        gsl_matrix_free(Qr);

        // compute X*transpose(Qnorm) to get topic dist in documents
        gsl_matrix * Xinf = gsl_matrix_alloc(N, K);
        // matrix multiply
        gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1.0, X, Q, 0.0, Xinf);
        // normalize
        normalize(Xinf, N, K, 2);
        writeSparseMatrix(outdir+"/X_inf.txt", Xinf, N, K);
        gsl_matrix_free(Xinf);

        std::ofstream falpha((outdir+"/alpha").c_str());
        falpha << alpha;
        falpha.close();

}


void getLogLikelihoodExistingClusters(long double * LL, const gsl_matrix * X, const gsl_matrix * Q, const double * sum_Q, size_t doc_i, size_t K, size_t M, double beta)
{

        for (size_t i = 0; i < K; i++) LL[i] = 0L;
        gsl_vector * q_z = gsl_vector_alloc(M);
        for (size_t z = 0; z < K; z++)
        {

                // q_z = Q(z,:) + X(doc_i,:);
                gsl_matrix_get_row (q_z, Q, z);
                gsl_vector_add(q_z, &(gsl_matrix_const_row(X, doc_i).vector));

                double sum_q_z = gsl_blas_dasum(q_z);

                LL[z] = (long double)(gsl_sf_lngamma(M*beta + sum_Q[z]) - gsl_sf_lngamma(M*beta + sum_q_z));

                // word_indeces = find(q_z ~= 0) --  skip for now
                for (size_t wi = 0; wi < M; wi++)
                {
                        if (gsl_vector_get(q_z, wi) == 0) continue;
                        LL[z] += (long double)(gsl_sf_lngamma(gsl_vector_get(q_z,wi) + beta) - gsl_sf_lngamma(gsl_matrix_get(Q, z, wi) + beta));
                }
        }
        gsl_vector_free(q_z);

        // dumpl(LL, K);

}

long double getLogLikelihoodNewCluster(const gsl_matrix * X, size_t doc_i, size_t M, double beta)
{
        long double L = 0L;
        double sum_q_z = gsl_blas_dasum(&(gsl_matrix_const_row(X, doc_i).vector));
        L += (long double)(gsl_sf_lngamma(M*beta) - gsl_sf_lngamma(M*beta + sum_q_z));
        // word_indeces = find(X(doc_i,:) ~= 0); -- skip for now
        for (size_t w = 0; w < M; w++)
        {
                if (gsl_matrix_get(X, doc_i, w) == 0) continue;
                L += (long double)(gsl_sf_lngamma(beta + gsl_matrix_get(X, doc_i, w)) - gsl_sf_lngamma(beta));
        }

        // printf("LL new = %Le\n", L);
        return L;
}


size_t sampleFromClusters(long double * LL_existing, long double LL_new, const gsl_matrix * clusters, size_t doc_i, size_t K, double alpha, const double * sum_clusters)
{

        // first calculate the posterior log likelihoods
        long double s_z_noi = 0L;
        for (size_t z = 0; z < K; z++)
        {
                // s_z_noi = (long double)(gsl_blas_dasum(&(gsl_matrix_const_row(clusters, z).vector))) - (long double)(gsl_matrix_get(clusters, z, doc_i));
                s_z_noi = (long double)(sum_clusters[z]) - (long double)(gsl_matrix_get(clusters, z, doc_i));

                LL_existing[z] += logl(s_z_noi); // - logl(N - 1 + alpha);
        }

        LL_new += logl((long double)alpha);      // - logl(N - 1 + alpha);

        // find min and max LL values
        long double maxL = -INFINITY;
        // long double minL = 0L;
        for (size_t z = 0; z < K; z++)
        {
                // if (LL_existing[z] < minL) minL = LL_existing[z];
                if (LL_existing[z] > maxL) maxL = LL_existing[z];
        }
        // if (LL_new < minL) minL = LL_new;
        if (LL_new > maxL) maxL = LL_new;

        // need special sampling trick to avoid normalization undeflow if min(p) < -11399
        //
        // c stores unnormalized cumulative sum
        long double * c = new long double[K+2];
        c[0] = 0L;
        long double u = 0L;
        for (size_t z = 0; z < K; z++)
        {
                u = expl(LL_existing[z] - maxL);
                c[z+1] = c[z] + u;
        }
        u = expl(LL_new - maxL);
        c[K+1] = c[K] + u;

        srand48(time(NULL));
        long double r = c[K+1] * (long double)(drand48());  // random double in the range [0.0, sum(u))
        for (size_t z = 0; z < K; z++)
        {
                if (c[z+1] > r)
                {
                        // we're in the interval [z,z+1], and this the sampled valued we want
                        delete [] c;
                        return z;
                }
        }
        delete [] c;
        return K;
}


void gammaln(const gsl_matrix * A, size_t nRows, size_t nCols, gsl_matrix * Out, double c)
{
        for (size_t i = 0; i < nRows; i++)
        {
                for (size_t j = 0; j < nCols; j++)
                {
                        gsl_matrix_set(Out, i, j, gsl_sf_lngamma(gsl_matrix_get(A, i, j) + c));
                }
        }
}


void gammaln(const gsl_vector * V, size_t dim, gsl_vector * Out, double c)
{
        for (size_t i = 0; i < dim; i++)
        {
                gsl_vector_set(Out, i, gsl_sf_lngamma(gsl_vector_get(V, i) + c));
        }
}


void normalize(gsl_matrix * A, size_t nRows, size_t nCols, size_t dim)
{
        if (dim == 1)
        {
                double * S = new double[nCols];
                sum(A, nRows, nCols, 1, S);
                for (size_t j = 0; j < nCols; j++)
                {
                        gsl_vector * _qcol = &(gsl_matrix_column(A, j).vector);
                        if (S[j] != 0)
                        {
                                gsl_vector_scale(_qcol, 1.0/S[j]);
                                gsl_matrix_set_col(A, j, _qcol);
                        }
                }
                delete [] S;
        }
        if (dim == 2)
        {
                double * S = new double[nRows];
                sum(A, nRows, nCols, 2, S);
                for (size_t i = 0; i < nRows; i++)
                {
                        gsl_vector * _qrow = &(gsl_matrix_row(A, i).vector);
                        if (S[i] != 0)
                        {
                                gsl_vector_scale(_qrow, 1.0/S[i]);
                                gsl_matrix_set_row(A, i, _qrow);
                        }
                }
                delete [] S;
        }
}


bool sum(const gsl_matrix * A, size_t nRows, size_t nCols, size_t dim, double * result)
{

        try
        {
                // sum rows (e.g. normalize each column)
                if (dim == 1)
                {
                        for (size_t j = 0; j < nCols; j++)
                        {
                                result[j] = 0;
                                for (size_t i = 0; i < nRows; i++) result[j] += gsl_matrix_get(A, i, j);
                        }
                        return true;
                }
                // sum cols (e.g normalize each row)
                else if (dim == 2)
                {
                        for (size_t i = 0; i < nRows; i++)
                        {
                                result[i] = 0;
                                for (size_t j = 0; j < nCols; j++) result[i] += gsl_matrix_get(A, i, j);
                        }
                        return true;
                }
                return false;
        }
        catch (...)
        {
                std::cerr << "Error in sum() \n";
                return false;
        }
}


bool readCSVMatrix(const std::string& filename, gsl_matrix * A, size_t nRows, size_t nCols)
{
        std::ifstream f;
        f.open(filename.c_str(), std::ios::in);
        if (!f.good()) return false;
        std::string line;
        std::vector<std::vector<double> > rows;
        std::vector<double> cols;
        size_t size1 = 0;
        size_t size2 = 0;
        while (!f.eof())
        {
                cols.clear();
                getline(f,line);
                if (line.size() == 0) break;
                boost::tokenizer<boost::escaped_list_separator<char> > tok(line);
                for(boost::tokenizer<boost::escaped_list_separator<char> >::iterator it = tok.begin(); it != tok.end(); ++it)
                {
                        cols.push_back(atof((*it).c_str()));
                }
                rows.push_back(cols);
                if (cols.size() > size2) size2 = cols.size();
        }
        f.close();
        size1 = rows.size();

        if (size1 == 0 || size2 == 0) return false;
        if (size1 != nRows || size2 != nCols) return false;

        gsl_matrix_set_zero(A);
        for (uint i = 0; i < rows.size(); i++)
        {
                uint j = 0;
                for (j = 0; j < rows[i].size(); j++)
                {
                        gsl_matrix_set(A, i, j, rows[i][j]);
                }
        }
        return true;

}

/*
* Read a text file into a sparse matrix A.  Matrix A is auto resized to match CVS file.
* Sparse matrix format is:
*
M_rows N_colums
row_index col_index:value col_index:value ....
row_index col_index:value col_index:value ....
....
*/
bool readSparseMatrix(const std::string& filename, gsl_matrix *& A, size_t& nRows, size_t& nCols)
{
        std::ifstream f;
        f.open(filename.c_str(), std::ios::in);
        if (!f.good()) return false;
        std::string line;
        getline(f,line);
        if (line.size() == 0) return false;
        std::string M_rows = line.substr(0, line.find(" "));
        std::string N_cols = line.substr(line.find(" ")+1);


        nRows = atoi(M_rows.c_str());
        nCols = atoi(N_cols.c_str());

        A = gsl_matrix_alloc(nRows, nCols);

        if (A == 0)
        {
                std::cerr << "Could not allocate matrix.\n";
                return false;
        }

        gsl_matrix_set_zero(A);

        while (!f.eof())
        {
                getline(f,line);
                if (line.size() == 0) break;
                std::string row_index = line.substr(0, line.find(" "));
                uint row = atoi(row_index.c_str());
                if (row < 0 || row > nRows) continue;
                line = line.substr(line.find(" ") + 1);
                boost::char_separator<char> sep(" ");
                boost::tokenizer<boost::char_separator<char> > tok(line, sep);
                for(boost::tokenizer<boost::char_separator<char> >::iterator it = tok.begin(); it != tok.end(); ++it)
                {
                        std::string col_index = (*it).substr(0, (*it).find(":"));
                        std::string value = (*it).substr((*it).find(":")+1);
                        uint col = atoi(col_index.c_str());
                        gsl_matrix_set(A, row, col, atof(value.c_str()));
                }
        }
        return true;
}



void dump(const gsl_vector * V, size_t dim)
{
        for (size_t i = 0; i < dim; i++) std::cout << gsl_vector_get(V,i) << "  ";
        std::cout << "\n";
}

void dump(const double * V, size_t dim)
{
        for (size_t i = 0; i < dim; i++) std::cout << V[i] << "  ";
        std::cout << "\n";
}

void dumpl(const long double * V, size_t dim)
{
        for (size_t i = 0; i < dim; i++)
        {
                printf("%Le  ", V[i]);
        }
        std::cout<< "\n";
}

void dump(const gsl_matrix * X, size_t N, size_t M)
{
        for (size_t i = 0; i < N; i++)
        {
                for (size_t j = 0; j < M; j++)
                {
                        std::cout << gsl_matrix_get (X, i, j) << ",";
                }
                std::cout << "\n";
        }
}

void ehandler (const char * reason, const char * file, int line, int gsl_errno)
{
        printf ("Error: %s\n", gsl_strerror (gsl_errno));
}
