/************************************************************************

(C) Copyright 2009, Vadim von Brzeski. vadimv@numberjack.net.

dp_sample is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dp_sample is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dp_sample.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <iostream>
#include <fstream>
#include <string>
#include <gflags/gflags.h>

DEFINE_int32(N, 100, "number of data points to generate; data matrix X will have N rows");
DEFINE_int32(M, 100, "number unique responses per datum; data matrix X will have M cols");
DEFINE_int32(W, 1000, "avg total number of responses per datum (drawn from a Poisson dist with mean W)");
DEFINE_string(outdir, "/tmp", "the pathname to the directory for output files");
DEFINE_double(alpha, 1.0, "dispersion parameter for Dirichlet Process; larger alpha implies greater number of clusters");
DEFINE_double(beta, 1.0, "hyperparameter for theta ~ G(0) = Dirichlet(beta), symmetric DP base distribution; larger beta implies more uniform (less peaked) multinomial distns over M responses");


int main(int argc, char** argv)
{

        if (argc < 2)
        {
                std::string usage = "Generates simulated data via a Dirichlet Process and the CRP.\n\n";
                usage += "Usage : \ndp_sample -N=<N data points> -M=<M unique responses> -W=<W avg num responses per datum> -outdir=<output-dir-path> [-alpha=<alpha>] [-beta=<beta>] \n\n";
                usage += "Outputs X_data.txt in the directory specified (the N x M data matrix).\n\n";
                usage += "\ndp_sample --helpshort explains the options in more detail\n\n";
                std::cerr << usage;
                google::SetUsageMessage("");
                google::ParseCommandLineFlags(&argc, &argv, true);
                exit(0);
        }

        google::ParseCommandLineFlags(&argc, &argv, true);

        // number of data points
        int N = FLAGS_N;
        // number of unique responses possible per datum
        int M = FLAGS_M;
        // avg total number of responses (draws) per sample (sampled from Poisson)
        int W = FLAGS_W;
        // alpha  = dispersion parameters for Dirichlet Process
        // larger alpha implies greater number of clusters (e.g. topics)
        double alpha = FLAGS_alpha;
        // beta = coefficient for DP base distribution G_0 = symmetric Dirichlet(beta)
        // larger beta implies less distinct (peaked) multinomial distributions over the M responses - less distinct topics
        // smaller beta implies more distinct (peaked) multinomial distn's over M responses - more distinct topics
        double beta = FLAGS_beta;

        std::string outdir = FLAGS_outdir;

        double * lambda = new double[M];
        for (size_t i = 0; i < M; i++) lambda[i] = beta;

        // setup random number generator
        const gsl_rng_type * T;
        gsl_rng * r;
        /* create a generator chosen by the environment variable GSL_RNG_TYPE */
        gsl_rng_env_setup();
        T = gsl_rng_default;
        r = gsl_rng_alloc (T);
        srand(time(NULL));

        // final Theta matrix
        gsl_matrix * Theta = gsl_matrix_alloc(1, M);
        size_t size_Theta = 1;

        // sample Theta[0,:]
        double * theta = new double[M];
        gsl_rng_set (r, rand());
        gsl_ran_dirichlet (r, M, lambda, theta);
        gsl_matrix_set_row (Theta, 0, &(gsl_vector_const_view_array (theta, M)).vector);
        double N_samples = 1;

        // keep track of cluster membership
        double * counts = new double[1];
        counts[0] = 1;

        for (size_t i = 2; i <= N; i++)
        {
                double prob_G_0 = alpha/(alpha + N_samples);
                double prob_existing = N_samples/(alpha + N_samples);
                gsl_rng_set (r, rand());
                int draw = gsl_ran_bernoulli (r, prob_G_0);
                if (draw == 1)
                {
                        // this is the sample from G_0 case
                        // std::cerr << "drawing from G_0\n";

                        // allocate temp matrix and copy over the previous matrix's values
                        gsl_matrix * _temp_Theta = gsl_matrix_alloc(size_Theta + 1, M);
                        for (size_t j = 0; j < size_Theta; j++)
                        {
                                gsl_matrix_set_row (_temp_Theta, j, &(gsl_matrix_const_row(Theta, j)).vector);
                        }

                        // draw from G_0
                        gsl_rng_set (r, rand());
                        gsl_ran_dirichlet (r, M, lambda, theta);
                        // set the last row of the (temp) matrix to the new draw
                        gsl_matrix_set_row (_temp_Theta, size_Theta, &(gsl_vector_const_view_array (theta, M)).vector);
                        gsl_matrix_free(Theta);
                        Theta = _temp_Theta;

                        // update counts
                        double * _temp_counts = new double[size_Theta + 1];
                        for (size_t j = 0; j < size_Theta; j++) _temp_counts[j] = counts[j];
                        _temp_counts[size_Theta] = 1;
                        delete [] counts;
                        counts = _temp_counts;

                        // update size (num rows) of Theta matrix
                        size_Theta += 1;
                }
                else
                {
                        // std::cerr << "drawing from existing samples\n";

                        // this is the sample from the existing samples
                        // uniformly draw from existing samples proportional to counts vector
                        unsigned int * sample = new unsigned int[size_Theta];
                        gsl_rng_set (r, rand());
                        gsl_ran_multinomial (r, size_Theta, 1, counts, sample);
                        size_t index = 0;
                        for (size_t j = 0; j < size_Theta; j++)
                        {
                                if (sample[j] == 1)
                                {
                                        index = j;
                                        break;
                                }
                        }
                        counts[index] += 1;
                }
                N_samples += 1;
        }

        std::cerr << "clusters : ";
        for (size_t i = 0; i < size_Theta; i++)
        {
                std::cerr << counts[i] << " ";
        }
        std::cerr << "\n";

        // the final data matrix X of samples
        gsl_matrix * X = gsl_matrix_alloc (N, M);
        gsl_matrix_set_zero(X);

        size_t i = 0;
        unsigned int * samples = new unsigned int[M];
        double * _d_samples = new double [M];
        for (size_t c = 0; c < size_Theta; c++)
        {
                int nInCluster = (int)counts[c];
                // get the theta vector of multinomial probs
                for (size_t j = 0; j < M; j++) theta[j] = gsl_matrix_get(Theta, c, j);

                for (size_t d = 0; d < nInCluster; d++)
                {
                        gsl_rng_set (r, rand());
                        // sample number of responses W from a Poisson with mean W
                        W = gsl_ran_poisson (r, W);
                        // sample based on the vector theta for this cluster
                        gsl_ran_multinomial (r, M, W, theta, samples);
                        for (size_t k = 0; k < M; k++) _d_samples[k] = samples[k];
                        gsl_matrix_set_row(X, i, &(gsl_vector_const_view_array (_d_samples, M)).vector);
                        i += 1;
                }
        }

        std::string filename = outdir + "/X_data.txt";
        std::ofstream outX(filename.c_str());
        outX << N << " " << M << "\n";
        for (size_t i = 0; i < N; i++)
        {
                outX << i << " ";
                for (size_t j = 0; j < M-1; j++)
                {
                        outX << j << ":" << gsl_matrix_get (X, i, j) << " ";
                }
                outX << (M-1) << ":" << gsl_matrix_get(X, i, M-1) << "\n";
        }
        outX.close();

        /*
        // write out binary format
        filename = outdir + "/X.dat";
        FILE * f = fopen (filename.c_str(), "wb");
        gsl_matrix_fwrite (f, X);
        fclose (f);
        */

        delete [] samples;
        delete [] _d_samples;
        delete [] theta;
        delete [] lambda;
        delete [] counts;
        gsl_matrix_free(Theta);

        return 0;
}

